@@ -52,7 +52,7 @@
 //! }
 //! ```
 //!
-//! ## Features
+//! ## Async runtimes and HTTP clients
 //!
 //! In order to support different async runtimes, the exporter requires you to specify an HTTP
 //! client that works with your chosen runtime. The [`opentelemetry-http`] crate comes with support
@@ -72,6 +72,46 @@
 //!
 //! Alternatively you can bring any other HTTP client by implementing the `HttpClient` trait.
 //!
+//! ## Metrics
+//!
+//! Please note: Metrics are still experimental both in the OpenTelemetry specification as well as
+//! Rust implementation.
+//!
+//! Please note: The metrics export configuration is still a bit rough in this crate. But once
+//! configured it should work as expected.
+//!
+//! This requires the **metrics** feature.
+//!
+//! ```no_run
+//! use opentelemetry::{global, sdk};
+//! use std::time::Duration;
+//!
+//! #[tokio::main]
+//! async fn main() {
+//!     // Setup exporter
+//!     let instrumentation_key = std::env::var("INSTRUMENTATION_KEY").unwrap();
+//!     let exporter = opentelemetry_application_insights::Exporter::new(instrumentation_key, ());
+//!     let controller = sdk::metrics::controllers::push(
+//!         sdk::metrics::selectors::simple::Selector::Exact,
+//!         sdk::export::metrics::ExportKindSelector::Stateless,
+//!         exporter,
+//!         tokio::spawn,
+//!         opentelemetry::util::tokio_interval_stream,
+//!     )
+//!     .with_period(Duration::from_secs(1))
+//!     .build();
+//!     global::set_meter_provider(controller.provider());
+//!
+//!     // Record value
+//!     let meter = global::meter("example");
+//!     let value_recorder = meter.f64_value_recorder("pi").init();
+//!     value_recorder.record(3.14, &[]);
+//!
+//!     // Give exporter some time to export values before exiting
+//!     tokio::time::sleep(Duration::from_secs(5)).await;
+//! }
+//! ```
+//!
 //! # Attribute mapping
 //!
 //! OpenTelemetry and Application Insights are using different terminology. This crate tries it's
@@ -147,39 +187,45 @@
 //! All other attributes are directly converted to custom properties.
 //!
 //! [exceptions]: https://github.com/open-telemetry/opentelemetry-specification/blob/master/specification/trace/semantic_conventions/exceptions.md
-#![doc(html_root_url = "https://docs.rs/opentelemetry-application-insights/0.17.0")]
+//!
+//! ## Metrics
+//!
+//! Metrics get reported to Application Insights as Metric Data. The [`Aggregator`] (chosen through
+//! the [`Selector`] passed to the controller) determines how the data is represented.
+//!
+//! | Aggregator     | Data representation                                       |
+//! | -------------- | --------------------------------------------------------- |
+//! | Array          | list of measurements                                      |
+//! | DDSketch       | aggregation with value, min, max and count                |
+//! | Histogram      | aggregation with sum and count (buckets are not exported) |
+//! | LastValue      | one measurement                                           |
+//! | MinMaxSumCount | aggregation with value, min, max and count                |
+//! | Sum            | aggregation with only a value                             |
+//!
+//! [`Aggregator`]: https://docs.rs/opentelemetry/0.16.0/opentelemetry/sdk/export/metrics/trait.Aggregator.html
+//! [`Selector`]: https://docs.rs/opentelemetry/0.16.0/opentelemetry/sdk/metrics/selectors/simple/enum.Selector.html
+#![doc(html_root_url = "https://docs.rs/opentelemetry-application-insights/0.18.0")]
 #![deny(missing_docs, unreachable_pub, missing_debug_implementations)]
+#![cfg_attr(docsrs, feature(doc_cfg))]
 #![cfg_attr(test, deny(warnings))]
 
 mod convert;
+#[cfg(feature = "metrics")]
+mod metrics;
 mod models;
 mod tags;
+mod trace;
 mod uploader;
 
-use async_trait::async_trait;
-use convert::{attrs_to_properties, duration_to_string, span_id_to_string, time_to_string};
 pub use models::context_tag_keys::attrs;
-use models::{
-    Data, Envelope, ExceptionData, ExceptionDetails, LimitedLenString1024, MessageData, Properties,
-    RemoteDependencyData, RequestData,
-};
 use opentelemetry::{
     global,
-    sdk::{
-        self,
-        export::{
-            trace::{ExportResult, SpanData, SpanExporter},
-            ExportError,
-        },
-        trace::TraceRuntime,
-    },
-    trace::{Event, SpanKind, StatusCode, TracerProvider},
-    Key, Value,
+    sdk::{self, export::ExportError, trace::TraceRuntime},
+    trace::TracerProvider,
 };
 pub use opentelemetry_http::HttpClient;
 use opentelemetry_semantic_conventions as semcov;
-use std::{borrow::Cow, collections::HashMap, convert::TryInto, error::Error as StdError};
-use tags::{get_tags_for_event, get_tags_for_span};
+use std::{borrow::Cow, convert::TryInto, error::Error as StdError};
 
 /// Create a new Application Insights exporter pipeline builder
 pub fn new_pipeline(instrumentation_key: String) -> PipelineBuilder<()> {
@@ -219,7 +265,7 @@ impl<C> PipelineBuilder<C> {
     /// Set endpoint used to ingest telemetry. This should consist of scheme and authrity. The
     /// exporter will call `/v2/track` on the specified endpoint.
     ///
-    /// Default: https://dc.services.visualstudio.com
+    /// Default: <https://dc.services.visualstudio.com>
     ///
     /// Note: This example requires [`reqwest`] and the **reqwest-client-blocking** feature.
     ///
@@ -425,7 +471,7 @@ impl<C> Exporter<C> {
     /// Set endpoint used to ingest telemetry. This should consist of scheme and authrity. The
     /// exporter will call `/v2/track` on the specified endpoint.
     ///
-    /// Default: https://dc.services.visualstudio.com
+    /// Default: <https://dc.services.visualstudio.com>
     pub fn with_endpoint(
         mut self,
         endpoint: &str,
@@ -443,78 +489,6 @@ impl<C> Exporter<C> {
         self.sample_rate = sample_rate * 100.0;
         self
     }
-
-    fn create_envelopes(&self, span: SpanData) -> Vec<Envelope> {
-        let mut result = Vec::with_capacity(1 + span.events.len());
-
-        let (data, tags, name) = match span.span_kind {
-            SpanKind::Server | SpanKind::Consumer => {
-                let data: RequestData = (&span).into();
-                let tags = get_tags_for_span(&span);
-                (
-                    Data::Request(data),
-                    tags,
-                    "Microsoft.ApplicationInsights.Request",
-                )
-            }
-            SpanKind::Client | SpanKind::Producer | SpanKind::Internal => {
-                let data: RemoteDependencyData = (&span).into();
-                let tags = get_tags_for_span(&span);
-                (
-                    Data::RemoteDependency(data),
-                    tags,
-                    "Microsoft.ApplicationInsights.RemoteDependency",
-                )
-            }
-        };
-        result.push(Envelope {
-            name: name.into(),
-            time: time_to_string(span.start_time).into(),
-            sample_rate: Some(self.sample_rate),
-            i_key: Some(self.instrumentation_key.clone().into()),
-            tags: Some(tags),
-            data: Some(data),
-        });
-
-        for event in span.events.iter() {
-            let (data, name) = match event.name.as_ref() {
-                "exception" => (
-                    Data::Exception(event.into()),
-                    "Microsoft.ApplicationInsights.Exception",
-                ),
-                _ => (
-                    Data::Message(event.into()),
-                    "Microsoft.ApplicationInsights.Message",
-                ),
-            };
-            result.push(Envelope {
-                name: name.into(),
-                time: time_to_string(event.timestamp).into(),
-                sample_rate: Some(self.sample_rate),
-                i_key: Some(self.instrumentation_key.clone().into()),
-                tags: Some(get_tags_for_event(&span)),
-                data: Some(data),
-            });
-        }
-
-        result
-    }
-}
-
-#[async_trait]
-impl<C> SpanExporter for Exporter<C>
-where
-    C: HttpClient,
-{
-    /// Export spans to Application Insights
-    async fn export(&mut self, batch: Vec<SpanData>) -> ExportResult {
-        let envelopes: Vec<_> = batch
-            .into_iter()
-            .flat_map(|span| self.create_envelopes(span))
-            .collect();
-
-        uploader::send(&self.client, &self.endpoint, envelopes).await
-    }
 }
 
 /// Errors that occurred during span export.
@@ -552,192 +526,3 @@ impl ExportError for Error {
         "application-insights"
     }
 }
-
-impl From<&SpanData> for RequestData {
-    fn from(span: &SpanData) -> RequestData {
-        let mut data = RequestData {
-            ver: 2,
-            id: span_id_to_string(span.span_context.span_id()).into(),
-            name: Some(LimitedLenString1024::from(span.name.clone()))
-                .filter(|x| !x.as_ref().is_empty()),
-            duration: duration_to_string(
-                span.end_time
-                    .duration_since(span.start_time)
-                    .unwrap_or_default(),
-            ),
-            response_code: (span.status_code as i32).to_string().into(),
-            success: span.status_code != StatusCode::Error,
-            source: None,
-            url: None,
-            properties: attrs_to_properties(&span.attributes, span.resource.clone()),
-        };
-
-        if let Some(method) = span.attributes.get(&semcov::trace::HTTP_METHOD) {
-            data.name = Some(
-                if let Some(route) = span.attributes.get(&semcov::trace::HTTP_ROUTE) {
-                    format!("{} {}", method.as_str(), route.as_str()).into()
-                } else {
-                    method.into()
-                },
-            );
-        }
-
-        if let Some(status_code) = span.attributes.get(&semcov::trace::HTTP_STATUS_CODE) {
-            data.response_code = status_code.into();
-        }
-
-        if let Some(url) = span.attributes.get(&semcov::trace::HTTP_URL) {
-            data.url = Some(url.into());
-        } else if let Some(target) = span.attributes.get(&semcov::trace::HTTP_TARGET) {
-            let mut target = target.as_str().into_owned();
-            if !target.starts_with('/') {
-                target.insert(0, '/');
-            }
-
-            if let (Some(scheme), Some(host)) = (
-                span.attributes.get(&semcov::trace::HTTP_SCHEME),
-                span.attributes.get(&semcov::trace::HTTP_HOST),
-            ) {
-                data.url =
-                    Some(format!("{}://{}{}", scheme.as_str(), host.as_str(), target).into());
-            } else {
-                data.url = Some(target.into());
-            }
-        }
-
-        if let Some(client_ip) = span.attributes.get(&semcov::trace::HTTP_CLIENT_IP) {
-            data.source = Some(client_ip.into());
-        } else if let Some(peer_ip) = span.attributes.get(&semcov::trace::NET_PEER_IP) {
-            data.source = Some(peer_ip.into());
-        }
-
-        data
-    }
-}
-
-impl From<&SpanData> for RemoteDependencyData {
-    fn from(span: &SpanData) -> RemoteDependencyData {
-        let mut data = RemoteDependencyData {
-            ver: 2,
-            id: Some(span_id_to_string(span.span_context.span_id()).into()),
-            name: span.name.clone().into(),
-            duration: duration_to_string(
-                span.end_time
-                    .duration_since(span.start_time)
-                    .unwrap_or_default(),
-            ),
-            result_code: Some((span.status_code as i32).to_string().into()),
-            success: match span.status_code {
-                StatusCode::Unset => None,
-                StatusCode::Ok => Some(true),
-                StatusCode::Error => Some(false),
-            },
-            data: None,
-            target: None,
-            type_: None,
-            properties: attrs_to_properties(&span.attributes, span.resource.clone()),
-        };
-
-        if let Some(status_code) = span.attributes.get(&semcov::trace::HTTP_STATUS_CODE) {
-            data.result_code = Some(status_code.into());
-        }
-
-        if let Some(url) = span.attributes.get(&semcov::trace::HTTP_URL) {
-            data.data = Some(url.into());
-        } else if let Some(statement) = span.attributes.get(&semcov::trace::DB_STATEMENT) {
-            data.data = Some(statement.into());
-        }
-
-        if let Some(host) = span.attributes.get(&semcov::trace::HTTP_HOST) {
-            data.target = Some(host.into());
-        } else if let Some(peer_name) = span.attributes.get(&semcov::trace::NET_PEER_NAME) {
-            if let Some(peer_port) = span.attributes.get(&semcov::trace::NET_PEER_PORT) {
-                data.target = Some(format!("{}:{}", peer_name.as_str(), peer_port.as_str()).into());
-            } else {
-                data.target = Some(peer_name.into());
-            }
-        } else if let Some(peer_ip) = span.attributes.get(&semcov::trace::NET_PEER_IP) {
-            if let Some(peer_port) = span.attributes.get(&semcov::trace::NET_PEER_PORT) {
-                data.target = Some(format!("{}:{}", peer_ip.as_str(), peer_port.as_str()).into());
-            } else {
-                data.target = Some(peer_ip.into());
-            }
-        } else if let Some(db_name) = span.attributes.get(&semcov::trace::DB_NAME) {
-            data.target = Some(db_name.into());
-        }
-
-        if span.span_kind == SpanKind::Internal {
-            data.type_ = Some("InProc".into());
-        } else if let Some(db_system) = span.attributes.get(&semcov::trace::DB_SYSTEM) {
-            data.type_ = Some(db_system.into());
-        } else if let Some(messaging_system) = span.attributes.get(&semcov::trace::MESSAGING_SYSTEM)
-        {
-            data.type_ = Some(messaging_system.into());
-        } else if let Some(rpc_system) = span.attributes.get(&semcov::trace::RPC_SYSTEM) {
-            data.type_ = Some(rpc_system.into());
-        } else if let Some(ref properties) = data.properties {
-            if properties.keys().any(|x| x.as_ref().starts_with("http.")) {
-                data.type_ = Some("HTTP".into());
-            } else if properties.keys().any(|x| x.as_ref().starts_with("db.")) {
-                data.type_ = Some("DB".into());
-            }
-        }
-
-        data
-    }
-}
-
-impl From<&Event> for ExceptionData {
-    fn from(event: &Event) -> ExceptionData {
-        let mut attrs: HashMap<&Key, &Value> = event
-            .attributes
-            .iter()
-            .map(|kv| (&kv.key, &kv.value))
-            .collect();
-        let exception = ExceptionDetails {
-            type_name: attrs
-                .remove(&semcov::trace::EXCEPTION_TYPE)
-                .map(Into::into)
-                .unwrap_or_else(|| "<no type>".into()),
-            message: attrs
-                .remove(&semcov::trace::EXCEPTION_MESSAGE)
-                .map(Into::into)
-                .unwrap_or_else(|| "<no message>".into()),
-            stack: attrs
-                .remove(&semcov::trace::EXCEPTION_STACKTRACE)
-                .map(Into::into),
-        };
-        ExceptionData {
-            ver: 2,
-            exceptions: vec![exception],
-            properties: Some(
-                attrs
-                    .iter()
-                    .map(|(k, v)| (k.as_str().into(), (*v).into()))
-                    .collect(),
-            )
-            .filter(|x: &Properties| !x.is_empty()),
-        }
-    }
-}
-
-impl From<&Event> for MessageData {
-    fn from(event: &Event) -> MessageData {
-        MessageData {
-            ver: 2,
-            message: if event.name.is_empty() {
-                "<no message>".into()
-            } else {
-                event.name.clone().into_owned().into()
-            },
-            properties: Some(
-                event
-                    .attributes
-                    .iter()
-                    .map(|kv| (kv.key.as_str().into(), (&kv.value).into()))
-                    .collect(),
-            )
-            .filter(|x: &Properties| !x.is_empty()),
-        }
-    }
-}
